# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path
from typing import Literal

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from anndata import AnnData

from cellarium.ml import CellariumModule
from cellarium.ml.data import (
    DistributedAnnDataCollection,
    IterableDistributedAnnDataCollectionDataset,
)
from cellarium.ml.utilities.data import AnnDataField, collate_fn
from tests.common import BoringModel

# RuntimeError: Too many open files. Communication with the workers is no longer possible.
# Please increase the limit using `ulimit -n` in the shell or change the sharing strategy
# by calling `torch.multiprocessing.set_sharing_strategy('file_system')` at the beginning of your code
torch.multiprocessing.set_sharing_strategy("file_system")


@pytest.fixture(params=[[3, 6, 9, 12], [4, 8, 12], [4, 8, 11]])  # limits
def dadc(tmp_path: Path, request: pytest.FixtureRequest):
    limits = request.param
    n_cell = limits[-1]
    g_gene = 1

    X = np.arange(n_cell).reshape(n_cell, g_gene)
    adata = AnnData(X, dtype=X.dtype)
    for i, limit in enumerate(zip([0] + limits, limits)):
        sliced_adata = adata[slice(*limit)]
        sliced_adata.write(os.path.join(tmp_path, f"adata.00{i}.h5ad"))

    # distributed anndata
    filenames = str(os.path.join(tmp_path, f"adata.{{000..00{len(limits)-1}}}.h5ad"))
    dadc = DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=3,
        cache_size_strictly_enforced=True,
    )
    return dadc


@pytest.mark.parametrize("iteration_strategy", ["same_order", "cache_efficient"])
@pytest.mark.parametrize("shuffle", [False, True], ids=["no shuffle", "shuffle"])
@pytest.mark.parametrize("num_workers", [0, 1, 2], ids=["zero workers", "one worker", "two workers"])
@pytest.mark.parametrize("batch_size", [1, 2, 3], ids=["batch size 1", "batch size 2", "batch size 3"])
def test_iterable_dataset(
    dadc: DistributedAnnDataCollection,
    iteration_strategy: Literal["same_order", "cache_efficient"],
    shuffle: bool,
    num_workers: int,
    batch_size: int,
):
    n_obs = len(dadc)
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        iteration_strategy=iteration_strategy,
        batch_keys={"x_ng": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    miss_counts = list(int(batch["miss_count"]) for batch in data_loader for _ in batch["x_ng"])
    actual_idx = list(int(i) for batch in data_loader for i in batch["x_ng"])

    if num_workers > 1:
        worker_ids = list(int(batch["worker_id"]) for batch in data_loader for _ in batch["x_ng"])
        adatas_oidx = np.searchsorted([0] + dadc.limits, actual_idx, side="right")
        for worker in range(num_workers):
            miss_count = max(c for c, w in zip(miss_counts, worker_ids) if w == worker)
            assert miss_count == len(set([o for o, w in zip(adatas_oidx, worker_ids) if w == worker]))
    else:
        miss_count = max(miss_counts)
        assert miss_count == len(dadc.limits)

    expected_idx = list(range(n_obs))

    # assert entire dataset is sampled
    if not shuffle and iteration_strategy == "same_order":
        assert expected_idx == actual_idx
    else:
        assert len(expected_idx) == len(actual_idx)
        assert set(expected_idx) == set(actual_idx)


@pytest.mark.parametrize("iteration_strategy", ["same_order", "cache_efficient"])
@pytest.mark.parametrize("shuffle", [False, True], ids=["no shuffle", "shuffle"])
@pytest.mark.parametrize("num_workers", [0, 1, 2], ids=["zero workers", "one worker", "two workers"])
@pytest.mark.parametrize("batch_size", [1, 2, 3], ids=["batch size 1", "batch size 2", "batch size 3"])
@pytest.mark.parametrize("drop_last", [False, True], ids=["no drop last", "drop last"])
def test_iterable_dataset_multi_device(
    dadc: DistributedAnnDataCollection,
    iteration_strategy: Literal["same_order", "cache_efficient"],
    shuffle: bool,
    num_workers: int,
    batch_size: int,
    drop_last: bool,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    n_obs = len(dadc)
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        iteration_strategy=iteration_strategy,
        batch_keys={"x_ng": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    # fit
    model = BoringModel()
    module = CellariumModule(model=model)
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=1,  # one pass
    )
    trainer.fit(module, data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    actual_idx = list(int(i) for batch in model.iter_data for i in batch["x_ng"])
    expected_idx = list(range(n_obs))

    # assert entire dataset is sampled
    if drop_last and n_obs % devices != 0:
        expected_len = (n_obs // devices) * devices
        assert expected_len == len(actual_idx)
        assert set(actual_idx).issubset(expected_idx)
    else:
        expected_len = math.ceil(n_obs / devices) * devices
        assert expected_len == len(actual_idx)
        assert set(expected_idx) == set(actual_idx)


@pytest.mark.parametrize(
    "num_workers,persistent_workers",
    [(0, False), (1, False), (1, True), (2, False), (2, True)],
    ids=[
        "zero workers",
        "one not persistent worker",
        "one persistent worker",
        "two not persistent workers",
        "two persistent workers",
    ],
)
@pytest.mark.parametrize("epochs", [2, 3], ids=["two epochs", "three epochs"])
def test_iterable_dataset_set_epoch_multi_device(
    dadc: DistributedAnnDataCollection,
    num_workers: int,
    persistent_workers: bool,
    epochs: int,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"x_ng": AnnDataField("X")},
        batch_size=1,
        shuffle=True,
        test_mode=True,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
        persistent_workers=persistent_workers,
    )

    # fit
    model = BoringModel()
    module = CellariumModule(model=model)
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=epochs,
        strategy="ddp",
    )
    trainer.fit(module, data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    iter_data = model.iter_data

    actual_epochs = set(int(i) for batch in iter_data for i in batch["epoch"])
    expected_epochs = set(range(epochs))

    assert set(expected_epochs) == set(actual_epochs)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


from typing import Literal

import lightning.pytorch as pl
import torch
from anndata import AnnData

from cellarium.ml.data import DistributedAnnDataCollection, IterableDistributedAnnDataCollectionDataset
from cellarium.ml.utilities.core import train_val_split
from cellarium.ml.utilities.data import AnnDataField, collate_fn


class CellariumAnnDataDataModule(pl.LightningDataModule):
    """
    DataModule for :class:`~cellarium.ml.data.IterableDistributedAnnDataCollectionDataset`.

    Example::

        >>> from cellarium.ml import CellariumAnnDataDataModule
        >>> from cellarium.ml.data import DistributedAnnDataCollection
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dm = CellariumAnnDataDataModule(
        ...     DistributedAnnDataCollection(
        ...         "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...         shard_size=10_000,
        ...     ),
        ...     max_cache_size=2,
        ...     batch_keys={
        ...         "x_ng": AnnDataField(attr="X", convert_fn=densify),
        ...         "var_names_g": AnnDataField(attr="var_names"),
        ...     },
        ...     batch_size=5000,
        ...     iteration_strategy="cache_efficient",
        ...     shuffle=True,
        ...     seed=0,
        ...     drop_last=True,
        ...     num_workers=4,
        ... )
        >>> dm.setup()
        >>> for batch in dm.train_dataloader():
        ...     print(batch.keys())  # x_ng, var_names_g

    Args:
        dadc:
            An instance of :class:`~cellarium.ml.data.DistributedAnnDataCollection` or :class:`AnnData`.
        batch_keys:
            Dictionary that specifies which attributes and keys of the :attr:`dadc` to return
            in the batch data and how to convert them. Keys must correspond to
            the input keys of the transforms or the model. Values must be instances of
            :class:`cellarium.ml.utilities.data.AnnDataField`.
        batch_size:
            How many samples per batch to load.
        iteration_strategy:
            Strategy to use for iterating through the dataset. Options are ``same_order`` and ``cache_efficient``.
            ``same_order`` will iterate through the dataset in the same order independent of the number of replicas
            and workers. ``cache_efficient`` will try to minimize the amount of anndata files fetched by each worker.
        shuffle:
            If ``True``, the data is reshuffled at every epoch.
        seed:
            Random seed used to shuffle the sampler if :attr:`shuffle=True`.
        drop_last:
            If ``True``, then the sampler will drop the tail of the data
            to make it evenly divisible across the number of replicas. If ``False``,
            the sampler will add extra indices to make the data evenly divisible across
            the replicas.
        train_size:
            Size of the train split. If :class:`float`, should be between ``0.0`` and ``1.0`` and represent
            the proportion of the dataset to include in the train split. If :class:`int`, represents
            the absolute number of train samples. If ``None``, the value is automatically set to the complement
            of the ``val_size``.
        val_size:
            Size of the validation split. If :class:`float`, should be between ``0.0`` and ``1.0`` and represent
            the proportion of the dataset to include in the validation split. If :class:`int`, represents
            the absolute number of validation samples. If ``None``, the value is set to the complement of
            the ``train_size``. If ``train_size`` is also ``None``, it will be set to ``0``.
        test_mode:
            If ``True`` enables tracking of cache and worker informations.
        num_workers:
            How many subprocesses to use for data loading. ``0`` means that the data will be loaded in the main process.
    """

    def __init__(
        self,
        dadc: DistributedAnnDataCollection | AnnData,
        # IterableDistributedAnnDataCollectionDataset args
        batch_keys: dict[str, AnnDataField] | None = None,
        batch_size: int = 1,
        iteration_strategy: Literal["same_order", "cache_efficient"] = "cache_efficient",
        shuffle: bool = False,
        seed: int = 0,
        drop_last: bool = False,
        train_size: float | int | None = None,
        val_size: float | int | None = None,
        test_mode: bool = False,
        # DataLoader args
        num_workers: int = 0,
    ) -> None:
        super().__init__()
        self.save_hyperparameters(logger=False)
        # Don't save dadc to the checkpoint
        self.hparams["dadc"] = None

        self.dadc = dadc
        # IterableDistributedAnnDataCollectionDataset args
        self.batch_keys = batch_keys or {}
        self.batch_size = batch_size
        self.iteration_strategy = iteration_strategy
        self.shuffle = shuffle
        self.seed = seed
        self.drop_last = drop_last
        self.n_train, self.n_val = train_val_split(len(dadc), train_size, val_size)
        self.test_mode = test_mode
        # DataLoader args
        self.num_workers = num_workers

    def setup(self, stage: str | None = None) -> None:
        """
        .. note::
           setup is called from every process across all the nodes. Setting state here is recommended.

        .. note::
            :attr:`val_dataset` is not shuffled and uses the ``same_order`` iteration strategy.

        """
        if stage == "fit":
            self.train_dataset = IterableDistributedAnnDataCollectionDataset(
                dadc=self.dadc,
                batch_keys=self.batch_keys,
                batch_size=self.batch_size,
                iteration_strategy=self.iteration_strategy,
                shuffle=self.shuffle,
                seed=self.seed,
                drop_last=self.drop_last,
                test_mode=self.test_mode,
                start_idx=0,
                end_idx=self.n_train,
            )
            self.val_dataset = IterableDistributedAnnDataCollectionDataset(
                dadc=self.dadc,
                batch_keys=self.batch_keys,
                batch_size=self.batch_size,
                iteration_strategy="same_order",
                shuffle=False,
                seed=self.seed,
                drop_last=False,
                test_mode=self.test_mode,
                start_idx=self.n_train,
                end_idx=self.n_train + self.n_val,
            )

        if stage == "predict":
            self.predict_dataset = IterableDistributedAnnDataCollectionDataset(
                dadc=self.dadc,
                batch_keys=self.batch_keys,
                batch_size=self.batch_size,
                iteration_strategy=self.iteration_strategy,
                shuffle=self.shuffle,
                seed=self.seed,
                drop_last=self.drop_last,
                test_mode=self.test_mode,
            )

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        """Training dataloader."""
        return torch.utils.data.DataLoader(
            self.train_dataset,
            num_workers=self.num_workers,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        """Validation dataloader."""
        return torch.utils.data.DataLoader(
            self.val_dataset,
            num_workers=self.num_workers,
            collate_fn=collate_fn,
        )

    def predict_dataloader(self) -> torch.utils.data.DataLoader:
        """Prediction dataloader."""
        return torch.utils.data.DataLoader(
            self.predict_dataset,
            num_workers=self.num_workers,
            collate_fn=collate_fn,
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


import torch
from torch import nn


class Log1p(nn.Module):
    """
    Log1p transform gene counts.

    .. math::

        y_{ng} = \\log(1 + x_{ng})
    """

    def __init__(self) -> None:
        super().__init__()

    def forward(self, x_ng: torch.Tensor) -> dict[str, torch.Tensor]:
        """
        .. note::

            When used with :class:`~cellarium.ml.core.CellariumModule` or :class:`~cellarium.ml.core.CellariumPipeline`,
            ``x_ng`` key in the input dictionary will be overwritten with the log1p transformed values.

        Args:
            x_ng: Gene counts.

        Returns:
            A dictionary with the following keys:

            - ``x_ng``: The log1p transformed gene counts.
        """
        x_ng = torch.log1p(x_ng)
        return {"x_ng": x_ng}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"
